/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is mozilla.org code.
 * 
 * The Initial Developers of the Original Code are Christopher
 * Blizzard and Jamie Zawinski.  Portions created Christopher Blizzard
 * are Copyright (C) 2000 Christopher Blizzard.  Portions created by
 * Jamie Zawinski are Copyright (C) 1994 Netscape Communications
 * Corporation.  All Rights Reserved.
 * 
 * Contributor(s): 
 */


#include "nsIGenericFactory.h"
#include "XRemoteClient.h"
#include <stdlib.h>
#include <unistd.h>
#include <sys/utsname.h>
#include <X11/Xatom.h>

#define MOZILLA_VERSION_PROP   "_MOZILLA_VERSION"
#define MOZILLA_LOCK_PROP      "_MOZILLA_LOCK"
#define MOZILLA_COMMAND_PROP   "_MOZILLA_COMMAND"
#define MOZILLA_RESPONSE_PROP  "_MOZILLA_RESPONSE"

#define XREMOTE_DEBUG

XRemoteClient::XRemoteClient()
{
  NS_INIT_REFCNT();
  mDisplay = 0;
  mInitialized = PR_FALSE;
  mMozVersionAtom = 0;
  mMozLockAtom = 0;
  mMozCommandAtom = 0;
  mMozResponseAtom = 0;
  mLockData = 0;
}

XRemoteClient::~XRemoteClient()
{
  if (mInitialized)
    Shutdown();
  if (mLockData)
    free(mLockData);
}

NS_IMPL_ADDREF(XRemoteClient)
NS_IMPL_RELEASE(XRemoteClient)

NS_INTERFACE_MAP_BEGIN(XRemoteClient)
  NS_INTERFACE_MAP_ENTRY(nsISupports)
  NS_INTERFACE_MAP_ENTRY(nsIXRemoteClient)
NS_INTERFACE_MAP_END

NS_IMETHODIMP
XRemoteClient::Init (void)
{
#ifdef XREMOTE_DEBUG
  printf("XRemoteClient::Init\n");
#endif

  if (mInitialized)
    return NS_OK;
  // try to open the display
  mDisplay = XOpenDisplay(0);
  if (!mDisplay)
    return NS_ERROR_FAILURE;

  // get our atoms
  mMozVersionAtom  = XInternAtom(mDisplay, MOZILLA_VERSION_PROP, False);
  mMozLockAtom     = XInternAtom(mDisplay, MOZILLA_LOCK_PROP, False);
  mMozCommandAtom  = XInternAtom(mDisplay, MOZILLA_COMMAND_PROP, False);
  mMozResponseAtom = XInternAtom(mDisplay, MOZILLA_RESPONSE_PROP, False);

  mInitialized = PR_TRUE;

  return NS_OK;
}

NS_IMETHODIMP
XRemoteClient::SendCommand (const char *aCommand, PRBool *aWindowFound)
{
#ifdef XREMOTE_DEBUG
  printf("XRemoteClient::SendCommand\n");
#endif

  *aWindowFound = PR_TRUE;

  // find the remote window
  Window window = FindWindow();

  // no window?  let the caller know.
  if (!window) {
    *aWindowFound = PR_FALSE;
    return NS_OK;
  }

  // make sure we get the right events on that window
  XSelectInput(mDisplay, window,
	       (PropertyChangeMask|StructureNotifyMask));
  
  nsresult rv;
  PRBool destroyed = PR_FALSE;

  // get the lock on the window
  rv = GetLock(window, &destroyed);

  if (NS_FAILED(rv))
    return NS_ERROR_FAILURE;

  // send our command
  rv = DoSendCommand(window, aCommand, &destroyed);

  // if the window was destroyed, don't bother trying to free the
  // lock.
  if (destroyed)
    return NS_ERROR_FAILURE;

  // doesn't really matter what this returns
  FreeLock(window);

  // if we failed above we had to free the lock - return the error
  // now.
  if (NS_FAILED(rv))
    return NS_ERROR_FAILURE;

  return NS_OK;
}

NS_IMETHODIMP
XRemoteClient::Shutdown (void)
{
#ifdef XREMOTE_DEBUG
  printf("XRemoteClient::Shutdown\n");
#endif

  if (!mInitialized)
    return NS_OK;
  // shut everything down
  XCloseDisplay(mDisplay);
  mDisplay = 0;
  mInitialized = PR_FALSE;
  return NS_OK;
}

Window
XRemoteClient::FindWindow(void)
{
  Window root = RootWindowOfScreen(DefaultScreenOfDisplay(mDisplay));
  Window root2, parent, *kids;
  unsigned int nkids;
  Window result = 0;
  int i;

  if (!XQueryTree(mDisplay, root, &root2, &parent, &kids, &nkids)) {
#ifdef XREMOTE_DEBUG
    NS_WARNING("XQueryTree failed in XRemoteClient::FindWindow");
#endif
    return 0;
  }

  if (!(kids && nkids)) {
#ifdef XREMOTE_DEBUG
    NS_WARNING("root window has no children");
#endif
    return 0;
  }

  for (i=nkids-1; i >= 0; i--) {
    Atom type;
    int format;
    unsigned long nitems, bytesafter;
    unsigned char *version = 0;
    Window w;
    w = kids[i];
    int status = XGetWindowProperty(mDisplay, w, mMozVersionAtom,
				    0, (65536 / sizeof (long)),
				    False, XA_STRING,
				    &type, &format, &nitems, &bytesafter,
				    &version);
    if (!version)
      continue;
    XFree(version);
    if (status == Success && type != None) {
      result = w;
      break;
    }
  }

#ifdef XREMOTE_DEBUG
  printf("FindWindow returning 0x%lx\n", result);
#endif

  return result;
}

nsresult
XRemoteClient::GetLock(Window aWindow, PRBool *aDestroyed)
{
  PRBool locked = PR_FALSE;
  PRBool waited = PR_FALSE;
  *aDestroyed = PR_FALSE;

  if (!mLockData) {
    mLockData = (char *)malloc(256);
    sprintf(mLockData, "pid%d@", getpid());
    struct utsname uname_result;
    int result;
    result = uname(&uname_result);
    if (result < 0) {
      NS_WARNING("failed to get hostname");
      return NS_ERROR_FAILURE;
    }
    strncat (mLockData, uname_result.nodename, 100);
  }

  do {
    int result;
    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *data = 0;

    XGrabServer(mDisplay);

    result = XGetWindowProperty (mDisplay, aWindow, mMozLockAtom,
				 0, (65536 / sizeof (long)),
				 False, /* don't delete */
				 XA_STRING,
				 &actual_type, &actual_format,
				 &nitems, &bytes_after,
				 &data);
    if (result != Success || actual_type == None) {
      /* It's not now locked - lock it. */
      XChangeProperty (mDisplay, aWindow, mMozLockAtom, XA_STRING, 8,
		       PropModeReplace, (unsigned char *) mLockData,
		       strlen (mLockData));
      locked = True;
    }

    XUngrabServer(mDisplay);
    XSync(mDisplay, False);

    if (!locked) {
      /* We tried to grab the lock this time, and failed because someone
	 else is holding it already.  So, wait for a PropertyDelete event
	 to come in, and try again. */
#ifdef XREMOTE_DEBUG
      fprintf (stderr, "window 0x%x is locked by %s; waiting...\n",
	       (unsigned int) aWindow, data);
#endif
      waited = True;
      while (1) {
	XEvent event;
	XNextEvent (mDisplay, &event);
	if (event.xany.type == DestroyNotify &&
	    event.xdestroywindow.window == aWindow) {
#ifdef XREMOTE_DEBUG
	  fprintf (stderr, "window 0x%x unexpectedly destroyed.\n",
		   (unsigned int) aWindow);
#endif
	  *aDestroyed = PR_TRUE;
	  return NS_ERROR_FAILURE;
	}
	else if (event.xany.type == PropertyNotify &&
		 event.xproperty.state == PropertyDelete &&
		 event.xproperty.window == aWindow &&
		 event.xproperty.atom == mMozLockAtom) {
	  /* Ok!  Someone deleted their lock, so now we can try
	     again. */
#ifdef XREMOTE_DEBUG
		  fprintf (stderr, "(0x%x unlocked, trying again...)\n",
			   (unsigned int) aWindow);
#endif
		  break;
	}
      }
    }
    if (data)
      XFree(data);
  } while (!locked);

#ifdef XREMOTE_DEBUG
  if (waited)
    fprintf(stderr, "obtained lock.\n");
#endif

  return NS_OK;
}

nsresult
XRemoteClient::FreeLock(Window aWindow)
{
  int result;
  Atom actual_type;
  int actual_format;
  unsigned long nitems, bytes_after;
  unsigned char *data = 0;

  result = XGetWindowProperty(mDisplay, aWindow, mMozLockAtom,
			      0, (65536 / sizeof(long)),
			      True, /* atomic delete after */
			      XA_STRING,
			      &actual_type, &actual_format,
			      &nitems, &bytes_after,
			      &data);
  if (result != Success) {
#ifdef XREMOTE_DEBUG
    printf("unable to read and delete " MOZILLA_LOCK_PROP
	   " property\n");
#endif
    return NS_ERROR_FAILURE;
  }
  else if (!data || !*data){
#ifdef XREMOTE_DEBUG
    fprintf (stderr, "invalid data on " MOZILLA_LOCK_PROP
	     " of window 0x%x.\n",
	     (unsigned int) aWindow);
#endif
    return NS_ERROR_FAILURE;
  }
  else if (strcmp((char *)data, mLockData)) {
#ifdef XREMOTE_DEBUG
    fprintf (stderr, MOZILLA_LOCK_PROP
	     " was stolen!  Expected \"%s\", saw \"%s\"!\n",
	     mLockData, data);
#endif
    return NS_ERROR_FAILURE;
  }

  if (data)
    XFree(data);
  return NS_OK;
}

nsresult
XRemoteClient::DoSendCommand(Window aWindow, const char *aCommand,
			     PRBool *aDestroyed)
{
  PRBool done = PR_FALSE;
  PRBool accepted = PR_FALSE;
  *aDestroyed = PR_FALSE;
  
#ifdef XREMOTE_DEBUG
  fprintf (stderr, "(writing " MOZILLA_COMMAND_PROP " \"%s\" to 0x%x)\n",
	   aCommand, (unsigned int) aWindow);
#endif

  XChangeProperty (mDisplay, aWindow, mMozCommandAtom, XA_STRING, 8,
		   PropModeReplace, (unsigned char *)aCommand,
		   strlen(aCommand));

  while (!done) {
    XEvent event;
    XNextEvent (mDisplay, &event);
    if (event.xany.type == DestroyNotify &&
	event.xdestroywindow.window == aWindow) {
      /* Print to warn user...*/
#ifdef XREMOTE_DEBUG
      fprintf (stderr, "window 0x%x was destroyed.\n",
	       (unsigned int) aWindow);
#endif
      *aDestroyed = PR_TRUE;
      goto DONE;
    }
    else if (event.xany.type == PropertyNotify &&
	     event.xproperty.state == PropertyNewValue &&
	     event.xproperty.window == aWindow &&
	     event.xproperty.atom == mMozResponseAtom) {
      Atom actual_type;
      int actual_format;
      unsigned long nitems, bytes_after;
      unsigned char *data = 0;
      Bool result;
      result = XGetWindowProperty (mDisplay, aWindow, mMozResponseAtom,
				   0, (65536 / sizeof (long)),
				   True, /* atomic delete after */
				   XA_STRING,
				   &actual_type, &actual_format,
				   &nitems, &bytes_after,
				   &data);
      if (result != Success) {
#ifdef XREMOTE_DEBUG
	fprintf (stderr, "failed reading " MOZILLA_RESPONSE_PROP
		 " from window 0x%0x.\n",
		 (unsigned int) aWindow);
#endif
	done = True;
      }
      else if (!data || strlen((char *) data) < 5) {
#ifdef XREMOTE_DEBUG
	fprintf (stderr, "invalid data on " MOZILLA_RESPONSE_PROP
		 " property of window 0x%0x.\n",
		 (unsigned int) aWindow);
#endif
	done = True;
      }
      else if (*data == '1') {	/* positive preliminary reply */
#ifdef XREMOTE_DEBUG
	fprintf (stderr, "%s\n", data + 4);
#endif
	/* keep going */
	done = False;
      }

      else if (!strncmp ((char *)data, "200", 3)) { /* positive completion */
	accepted = PR_TRUE;
	done = True;
      }

      else if (*data == '2') {	/* positive completion */
#ifdef XREMOTE_DEBUG
	fprintf (stderr, "%s\n", data + 4);
#endif
	accepted = PR_TRUE;
	done = True;
      }

      else if (*data == '3') {	/* positive intermediate reply */
#ifdef XREMOTE_DEBUG
	fprintf (stderr, "internal error: "
		 "server wants more information?  (%s)\n",
		 data);
#endif
	done = True;
      }

      else if (*data == '4' ||	/* transient negative completion */
	       *data == '5') {	/* permanent negative completion */
#ifdef XREMOTE_DEBUG
	fprintf (stderr, "%s\n", data + 4);
#endif
	done = True;
      }

      else {
#ifdef XREMOTE_DEBUG
	fprintf (stderr,
		 "unrecognised " MOZILLA_RESPONSE_PROP
		 " from window 0x%x: %s\n",
		 (unsigned int) aWindow, data);
#endif
	done = True;
      }

      if (data)
	XFree(data);
    }

    else if (event.xany.type == PropertyNotify &&
	     event.xproperty.window == aWindow &&
	       event.xproperty.state == PropertyDelete &&
	       event.xproperty.atom == mMozCommandAtom) {
#ifdef XREMOTE_DEBUG
      fprintf (stderr, "(server 0x%x has accepted "
	       MOZILLA_COMMAND_PROP ".)\n",
	       (unsigned int) aWindow);
#endif
    }
    
  }

 DONE:

  if (!accepted)
    return NS_ERROR_FAILURE;
  
  return NS_OK;
}

NS_GENERIC_FACTORY_CONSTRUCTOR(XRemoteClient)

static nsModuleComponentInfo components[] =
{
  { "XRemote Client",
    NS_XREMOTECLIENT_CID,
    "mozilla.xremote_client.1",
    XRemoteClientConstructor }
};

NS_IMPL_NSGETMODULE("XRemoteClientModule", components);
  
